// This file was generated by ecm_setup_version(): DO NOT EDIT!

#ifndef QACCESSIBILITYCLIENT_VERSION_H
#define QACCESSIBILITYCLIENT_VERSION_H

#define QACCESSIBILITYCLIENT_VERSION_STRING "0.4.1"
#define QACCESSIBILITYCLIENT_VERSION_MAJOR 0
#define QACCESSIBILITYCLIENT_VERSION_MINOR 4
#define QACCESSIBILITYCLIENT_VERSION_PATCH 1
#define QACCESSIBILITYCLIENT_VERSION ((0<<16)|(4<<8)|(1))

#endif
